#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "uint128_t.h"

// #define DEBUG_LOG

#define HI(n) ((uint64_t)(n)>>32)
#define LO(n) ((uint64_t)(n)&0xffffffff)
#define U128(high,low) ((uint128_t)(uint64_t)high<<64 | (uint64_t)low)

//---------------------------------------------------------------------------
// 0.45.0 - known values
//---------------------------------------------------------------------------

uint64_t magic_table[16] = {
  0x2dd7caaefcf073eb, 0xa9209937349cfe9c,
  0xb84bfc934b0e60ef, 0xff709c157b26e477,
  0x3936fd8735455112, 0xca141bf22338d331,
  0xdd40e749cb64fd02, 0x5e268f564b0deb26,
  0x658239596bdea9ec, 0x31cedf33ac38c624,
  0x12f56816481b0cfd, 0x94e9de155f40f095,
  0x5089c907844c6325, 0xdf887e97d73c50e3,
  0xae8870787ce3c11d, 0xa6767d18c58d2117,
};
#define ROUND_MAGIC_HI 0xe3f0d44988bcdfab
#define ROUND_MAGIC_LO 0x081570afdd535ec3
#define FINAL_MAGIC0 0xce7c4801d683e824
#define FINAL_MAGIC1 0x6823775b1daad522
#define ROUND_MAGIC U128(ROUND_MAGIC_HI,ROUND_MAGIC_LO)
// -- end

//---------------------------------------------------------------------------

// PHASE 1: 
// - find the 8 combinations for ROUND_MAGIC, FINAL_MAGIC0 and FINAL_MAGIC1
uint32_t magic_candidates[61]
{
0x54850980, 0x35764bce, 0xed368682, 0x6fccf42c, 0x2019fe30, 0xaf286bcb,
0xaaaaaaab, 0xaf286bcb, 0x24924925, 0xaaaaaaab, 0x5ccb1c71, 0x5ccb1c71,
0xe3f0d449, 0x88bcdfab, 0xe3f070af, 0x88bc5ec3, 0x081570af, 0xdd535ec3,
0x656e772b, 0x5a6ba2df, 0x656e772b, 0x5a6ba2df, 0x478d176d, 0x609d6e72,
0x5a6ba2df, 0x5a6b772b, 0x656e772b, 0x138b1f1f, 0xaaaaaaab, 0x5908ad45,
0xca4587e7, 0xaaaaaaab, 0x45bafc91, 0x16eafc99, 0xd2784bb1, 0xcee1b01e,
0x8844d71d, 0xb7719b7d, 0xb4342ed9, 0x5f6e661e, 0x2a6cabbf, 0xd683e824,
0xce7c4801, 0x297c17db, 0x297c38e2, 0x3183b7fe, 0x8e4338e2, 0x1daad522,
0x6823775b, 0xe2552add, 0x97dc88a4, 0xaa9b8763, 0x68bd30fe, 0x06f68d33,
0xaaaaaaab, 0x76ecb897, 0xd8d90d05, 0xda6ff4e6, 0x01b8a6cc, 0xfffffefe,
0xc7a2ff89 
}; 
int       magic_candidate_count = 61; 

// global iterators
int g_i0, g_i1, g_i2, g_i3, g_i4, g_i5, g_i6, g_i7;

// functions for brute-force
uint64_t get_ROUND_MAGIC_HI()
{
  return ((uint64_t)magic_candidates[g_i1] << 32) | magic_candidates[g_i0];
}

uint64_t get_ROUND_MAGIC_LO()
{
  return ((uint64_t)magic_candidates[g_i3] << 32) | magic_candidates[g_i2];
}

uint64_t get_FINAL_MAGIC0()
{
  return ((uint64_t)magic_candidates[g_i5] << 32) | magic_candidates[g_i4];
}

uint64_t get_FINAL_MAGIC1()
{
  return ((uint64_t)magic_candidates[g_i7] << 32) | magic_candidates[g_i6];
}

uint128_t get_ROUND_MAGIC()
{
  return U128(get_ROUND_MAGIC_HI(),get_ROUND_MAGIC_LO());
}

// END
//---------------------------------------------------------------------------

uint64_t Hash(uint8_t *in, uint32_t len)
{
  uint128_t hash = 0;

  hash = get_ROUND_MAGIC();

  // Finalize the hash
  hash += U128(0, 0);
  if (hash >= U128(0x7fffffffffffffff,0xffffffffffffffff)) {
    hash++;
  }
  hash = hash << 1 >> 1;

  uint64_t hash_high = hash >> 64;
  uint64_t hash_low = hash;
  uint64_t X = hash_high + HI(hash_low);
  X = HI(X + HI(X) + 1) + hash_high;
  uint64_t Y = (X << 32) + hash_low;

  uint64_t A = X + get_FINAL_MAGIC0();
  if (A < X) {
    A += 0x101;
  }

  uint64_t B = Y + get_FINAL_MAGIC1();
  if (B < Y) {
    B += 0x101;
  }

  uint128_t tH;
  uint128_t H = (uint128_t) A * B;
  tH = (H >> 64); tH *= 0x101; tH += (uint64_t) H; H = tH;
  tH = (H >> 64); tH *= 0x101; tH += (uint64_t) H; H = tH;
  if (H >> 64) {
    H += 0x101;
  }
  if ((uint64_t) H > 0xFFFFFFFFFFFFFEFE) {
    H += 0x101;
  }
  return (uint64_t) H;
}

int main(int argc, char **argv)
{
  FILE *f;
  char  line[64], *p, *_p;
  int   i, min, max;
  int   attempts = 0;

  fprintf(stdout, "PogoHash-brute #1\n");
  fprintf(stdout, "-----------------\n");
  fprintf(stdout, "\n");

  // lets find 0xfffffefe : our magic number
  min = 0;
  max = magic_candidate_count-1;
  while ((max > 8) && (magic_candidates[max] != 0xfffffefe)) max--;

  // we know that the value isn't a magic
  max--;
 
  // diff between x.len_0.load and x.len_1.load happens at index 12
  max = 48;
  min = 12;

  fprintf(stdout, "x.candidates\n");
  fprintf(stdout, "------------\n");
  fprintf(stdout, "  total: %d\n", magic_candidate_count);
  fprintf(stdout, "  min:   %d\n", min);
  fprintf(stdout, "  max:   %d\n", max);
  fprintf(stdout, "\n");
 
  fprintf(stdout, ":: Hash(buffer, sizeof(buffer) [iOS code]\n");
  {
    uint64_t result;
    uint8_t  buffer[256] = {0};
    int      j;

    // this is our test vector
    j = 0;

    //
    // BRUTE FORCE
    //

    fprintf(stdout, "scanning:\n");

    // FINAL_MAGIC1
    g_i7 = max;
loop_i7:;
    {
      fprintf(stdout, "g_i7=%d\n", g_i7);

      g_i6 = max;
loop_i6:;
      {
        // we must find the lowest of the two indexes
        int t1 = (g_i6 > g_i7) ? g_i7 : g_i6;

        // special case: cannot happen
        if (g_i6 == g_i7) goto loop_i6_done; 

        // FINAL_MAGIC0
        g_i5=t1-1;
loop_i5:;
        {
          g_i4=t1-1;
loop_i4:;
          {
            // we must find the lowest of the two indexes
            int t2 = (g_i4 > g_i5) ? g_i5 : g_i4;

            // special case: cannot happen
            if (g_i4 == g_i5) goto loop_i4_done;

            // ROUND_MAGIC
            g_i3=t2-1; 
loop_i3:; 
            {
              g_i2=t2-1;
loop_i2:;
              {
                // special case: cannot happen
                if (g_i2 == g_i3) goto loop_i2_done;

                g_i1=t2-1;
loop_i1:;
                {
                  // special case: cannot happen
                  if ((g_i1 == g_i2) || (g_i1 == g_i3)) goto loop_i1_done;

                  g_i0=t2-1;
loop_i0:;
                  {
                    // special case: cannot happen
                    if ((g_i0 == g_i1) || (g_i0 == g_i2) || (g_i0 == g_i3))
                      goto loop_i0_done;

                    attempts++;
                    result = Hash(buffer, j);
#ifdef DEBUG_LOG
                    fprintf(stdout, ":: %2d %2d %2d %2d %2d %2d %2d %2d\n", 
                            g_i0, g_i1, g_i2, g_i3, g_i4, g_i5, g_i6, g_i7);
#endif

                    // have we found them?
                    if (result == 0xeebb9c527a4d5f53) 
                    {
                      fprintf(stdout, ":: %2d %2d %2d %2d %2d %2d %2d %2d\n", 
                              g_i0, g_i1, g_i2, g_i3, g_i4, g_i5, g_i6, g_i7);

                      fprintf(stdout, "ROUND_MAGIC_HI = 0x%08x%08x\n",
                                      magic_candidates[g_i1], magic_candidates[g_i0]);
                      fprintf(stdout, "ROUND_MAGIC_LO = 0x%08x%08x\n",
                                      magic_candidates[g_i3], magic_candidates[g_i2]);
                      fprintf(stdout, "FINAL_MAGIC0   = 0x%08x%08x\n",
                                      magic_candidates[g_i5], magic_candidates[g_i4]);
                      fprintf(stdout, "FINAL_MAGIC1   = 0x%08x%08x\n",
                                      magic_candidates[g_i7], magic_candidates[g_i6]);

                      goto MAGICS_FOUND;
                    }
loop_i0_done:;
                    g_i0--;
                  }
                  if (g_i0>=min) goto loop_i0;

loop_i1_done:;
                  g_i1--;
                }
                if (g_i1>=min) goto loop_i1;

loop_i2_done:;
                g_i2--;
              }
              if (g_i2>=min) goto loop_i2;

              g_i3--;
            } 
            if (g_i3>=min) goto loop_i3;
          
loop_i4_done:;
            g_i4--;
          }
          if (g_i4 >= (min+2)) goto loop_i4;

          g_i5--;
        }
        if (g_i5 >= (min+2)) goto loop_i5;

loop_i6_done:;
        g_i6--;
      } 
      if (g_i6 >= (min+6)) goto loop_i6;

      g_i7--;
    } 
    if (g_i7 >= (min+6)) goto loop_i7;
  }
  fprintf(stdout, "NOT FOUND\n");
  goto MAGICS_NOT_FOUND;

MAGICS_FOUND:;

  // display how many child spawns were generated
  fprintf(stdout, "attempts: %d\n", attempts);

MAGICS_NOT_FOUND:

  fprintf(stdout, "done\n");
  return 0;
}
